# coding: utf-8
###
 # @file   reproduce_cifar.py
 # @author John Stephan <john.stephan@epfl.ch>
 #
 # @section LICENSE
 #
 # Copyright © 2021-2022 École Polytechnique Fédérale de Lausanne (EPFL).
 # All rights reserved.
 #
 # @section DESCRIPTION
 #
 # Running experiments and plots for CIFAR-10 dataset.
###

import tools
tools.success("Module loading...")
import argparse
import math
import pathlib
import signal
import shlex
import sys
import torch
import experiments

# ---------------------------------------------------------------------------- #
# Miscellaneous initializations
tools.success("Miscellaneous initializations...")

# "Exit requested" global variable accessors
exit_is_requested, exit_set_requested = tools.onetime("exit")

# Signal handlers
signal.signal(signal.SIGINT, exit_set_requested)
signal.signal(signal.SIGTERM, exit_set_requested)

# ---------------------------------------------------------------------------- #
# Command-line processing
tools.success("Command-line processing...")

def process_commandline():
  """ Parse the command-line and perform checks.
  Returns:
    Parsed configuration
  """
  # Description
  parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
  parser.add_argument("--result-directory",
    type=str,
    default="results-data",
    help="Path of the data directory, containing the data gathered from the experiments")
  parser.add_argument("--plot-directory",
    type=str,
    default="results-plot",
    help="Path of the plot directory, containing the graphs traced from the experiments")
  parser.add_argument("--devices",
    type=str,
    default="auto",
    help="Comma-separated list of devices on which to run the experiments, used in a round-robin fashion")
  parser.add_argument("--supercharge",
    type=int,
    default=1,
    help="How many experiments are run in parallel per device, must be positive")
  # Parse command line
  return parser.parse_args(sys.argv[1:])

with tools.Context("cmdline", "info"):
  args = process_commandline()
  # Check the "supercharge" parameter
  if args.supercharge < 1:
    tools.fatal(f"Expected a positive supercharge value, got {args.supercharge}")
  # Make the result directories
  def check_make_dir(path):
    path = pathlib.Path(path)
    if path.exists():
      if not path.is_dir():
        tools.fatal(f"Given path {str(path)!r} must point to a directory")
    else:
      path.mkdir(mode=0o755, parents=True)
    return path
  args.result_directory = check_make_dir(args.result_directory)
  args.plot_directory = check_make_dir(args.plot_directory)
  # Preprocess/resolve the devices to use
  if args.devices == "auto":
    if torch.cuda.is_available():
      args.devices = list(f"cuda:{i}" for i in range(torch.cuda.device_count()))
    else:
      args.devices = ["cpu"]
  else:
    args.devices = list(name.strip() for name in args.devices.split(","))

# ---------------------------------------------------------------------------- #
# Serial preloading of the dataset
tools.success("Pre-downloading datasets...")

# Pre-load the datasets to prevent the first parallel runs from downloading them several times
with tools.Context("dataset", "info"):
  for name in ("cifar10", "cifar100"):
    with tools.Context(name, "info"):
      experiments.make_datasets(name, 1, 1)

# ---------------------------------------------------------------------------- #
# Run (missing) experiments
tools.success("Running experiments...")

# Hyperparameters to test
byzcounts = [5, 11]
gars = ("brute", "cge", "krumStar", "median", "trmean", "centeredclip", "rfa", "meamed")
attacks = (("little", "factor:1.5 negative:True"), ("empire", "factor:1.1"), ("signflipping", None), ("labelflipping", None))
momentums = [0.0, 0.6, 0.8, 0.9, 0.99, 0.999]
learning_rates = [(0.25, 1500)]
datasets = [("cifar10", "cifar100:False")]

# Command maker helper
def make_command(params):
  cmd = ["python3", "-OO", "momentum.py"]
  cmd += tools.dict_to_cmdlist(params)
  return tools.Command(cmd)

# Jobs
jobs  = tools.Jobs(args.result_directory, devices=args.devices, devmult=args.supercharge)
seeds = jobs.get_seeds()

# Base parameters for the CIFAR-10 experiments
params_cifar10 = {
  "batch-size": 50,
  "model": "empire-cnn",
  "loss": "nll",
  "learning-rate-decay": 167,
  "momentum-at": "worker",
  "l2-regularize": 1e-2,
  "evaluation-delta": 100,
  "gradient-clip": 5,
  "gradient-clip-centered": 5,
  "nb-steps": 2500,
  "nb-for-study": 1,
  "nb-for-study-past": 25,
  "nb-workers": 25,
  "momentum-at": "worker"
}

# Submit all CIFAR-10/100 experiments
for ds, mp in datasets:
  for f in byzcounts:
    for lr, dd in learning_rates:
      for momentum in momentums:
        # No attack
        params = params_cifar10.copy()
        params["dataset"] = ds
        params["model-args"] = mp
        params["nb-workers"] -= f
        params["learning-rate"] = lr
        params["learning-rate-decay-delta"] = dd
        params["momentum"] = params["dampening"] = momentum
        jobs.submit(f"{ds}-average-n_{params['nb-workers']}-m_{momentum}-lr_{lr}", make_command(params))
        # Attacked
        for gar in gars:
            if f == 11 and gar == "brute":
                continue
            for attack, attargs in attacks:
                params = params_cifar10.copy()
                params["dataset"] = ds
                params["model-args"] = mp
                params["learning-rate"] = lr
                if gar == "cge":
                    params["learning-rate"] /= params["nb-workers"]
                params["learning-rate-decay-delta"] = dd
                params["nb-decl-byz"] = params["nb-real-byz"] = f
                params["gar"] = gar
                params["momentum"] = params["dampening"] = momentum
                params["attack"] = attack
                if attack == "labelflipping":
                    params["flip"] = True
                if attargs is not None:
                    params["attack-args"] = shlex.split(attargs)
                jobs.submit(f"{ds}-{attack}-{gar}-f_{f}-m_{momentum}-lr_{lr}", make_command(params))

# Wait for the jobs to finish and close the pool
jobs.wait(exit_is_requested)
jobs.close()

# Check if exit requested before going to plotting the results
if exit_is_requested():
  exit(0)

# ---------------------------------------------------------------------------- #
# Plot results
tools.success("Plotting results...")

# Import additional modules
try:
  import numpy
  import pandas
  import study
except ImportError as err:
  tools.fatal(f"Unable to plot results: {err}")

# GAR renaming
gar_to_name = {"brute": "MDA", "cge": "CGE", "krumStar": "Krum*", "median": "CWMED", "trmean": "CWTM", "centeredclip": "CC", "rfa": "GM", "meamed": "MeaMed"}

def compute_avg_err_op(name, *colops, avgs="", errs="-err"):
    """ Compute the average and standard deviation of the selected columns over the given experiment.
    Args:
      name Given experiment name
      ...  Tuples of (selected column name (through 'study.select'), optional reduction operation name)
      avgs Suffix for average column names
      errs Suffix for standard deviation (or "error") column names
    Returns:
      Data frames for each of the computed columns,
      Tuple of reduced values per seed (or None if None was provided for 'op')
    Raises:
      'RuntimeError' if a reduction operation was specified for a column selector that did not select exactly 1 column
    """
    # Load all the runs for the given experiment name, and keep only a subset
    datas = tuple(study.select(study.Session(args.result_directory / f"{name}-{seed}").compute_ratio(nowarn=True), *(col for col, _ in colops)) for seed in seeds)
    # Make the aggregated data frames
    def make_df_ro(col, op):
      nonlocal datas
      # For every selected columns
      subds = tuple(study.select(data, col).dropna() for data in datas)
      df    = pandas.DataFrame(index=subds[0].index)
      ro    = None
      for cn in subds[0]:
        # Generate compound column names
        avgn = cn + avgs
        errn = cn + errs
        # Compute compound columns
        numds = numpy.stack(tuple(subd[cn].to_numpy() for subd in subds))
        df[avgn] = numds.mean(axis=0)
        df[errn] = numds.std(axis=0)
        # Compute reduction, if requested
        if op is not None:
          if ro is not None:
            raise RuntimeError(f"column selector {col!r} selected more than one column ({(', ').join(subds[0].columns)}) while a reduction operation was requested")
          ro = tuple(getattr(subd[cn], op)().item() for subd in subds)
      # Return the built data frame and optional computed reduction
      return df, ro
    dfs = list()
    ros = list()
    for col, op in colops:
      df, ro = make_df_ro(col, op)
      dfs.append(df)
      ros.append(ro)
    # Return the built data frames and optional computed reductions
    return dfs, ros

def compute_mean_dev(vals):
    """ Compute the mean and standard deviation of an iterable of values.
    Args:
      vals Iterable of values
    Returns:
      Mean value,
      Standard deviation
    Raises:
      'AssertionError' if less than two values
    """
    assert len(vals) >= 2
    mean = sum(vals) / len(vals)
    sdev = math.sqrt(sum((val - mean) ** 2 for val in vals))
    return mean, sdev

# Plot (Fashion-)MNIST results
with tools.Context(ds, "info"):
    for ds, _ in datasets:
      maxaccs = dict()
      for momentum in momentums:
        for lr, _ in learning_rates:
         for f in byzcounts:
            # No attack
            name = f"{ds}-average-n_{params_cifar10['nb-workers'] - f}-m_{momentum}-lr_{lr}"
            try:
              noattack, ros = compute_avg_err_op(name, ("Accuracy", "max"))
              maxaccs[(momentum, lr, f)] = compute_mean_dev(ros[0])
            except Exception as err:
              tools.warning(f"Unable to process {name}: {err}")
              continue
            # Attack
            for attack, _ in attacks:
              name = f"{ds}-{attack}-%s-f_{f}-m_{momentum}-lr_{lr}"
              attacked = dict()
              for gar in gars:
                try:
                  cols, ros = compute_avg_err_op(name % gar, ("Accuracy", "max"))
                  attacked[gar] = cols
                  maxaccs[(momentum, lr, f, attack, gar)] = compute_mean_dev(ros[0])
                except Exception as err:
                  tools.warning(f"Unable to process {name % gar!r}: {err}")
                  continue
              # Plot top-1 cross-accuracy
              plot = study.LinePlot()
              plot.include(noattack[0], "Accuracy", errs="-err", lalp=0.8)
              legend = ["No attack"]
              for gar in gars:
                if gar not in attacked:
                  continue
                plot.include(attacked[gar][0], "Accuracy", errs="-err", lalp=0.8)
                legend.append(gar_to_name.get(gar, gar.capitalize()))
              #plot every time graph in terms of the maximum number of steps
              plot.finalize(None, "Step number", "Top-1 cross-accuracy", xmin=0, xmax=2500, ymin=0, ymax=1, legend=legend)
              plot.save(args.plot_directory / f"{ds}-{attack}-f_{f}-m_{momentum}-lr_{lr}.pdf", xsize=3, ysize=1.5)
